/*
 * CDE - Common Desktop Environment
 *
 * Copyright (c) 1993-2012, The Open Group. All rights reserved.
 *
 * These libraries and programs are free software; you can
 * redistribute them and/or modify them under the terms of the GNU
 * Lesser General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * These libraries and programs are distributed in the hope that
 * they will be useful, but WITHOUT ANY WARRANTY; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with these libraries and programs; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */
/* $XConsortium: fastfind.c /main/3 1995/11/01 17:55:25 rswiston $ */
/***************************************************************
*                                                              *
*                      AT&T - PROPRIETARY                      *
*                                                              *
*         THIS IS PROPRIETARY SOURCE CODE LICENSED BY          *
*                          AT&T CORP.                          *
*                                                              *
*                Copyright (c) 1995 AT&T Corp.                 *
*                     All Rights Reserved                      *
*                                                              *
*           This software is licensed by AT&T Corp.            *
*       under the terms and conditions of the license in       *
*       http://www.research.att.com/orgs/ssr/book/reuse        *
*                                                              *
*               This software was created by the               *
*           Software Engineering Research Department           *
*                    AT&T Bell Laboratories                    *
*                                                              *
*               For further information contact                *
*                     gsf@research.att.com                     *
*                                                              *
***************************************************************/

/* : : generated by proto : : */

#if !defined(__PROTO__)
#if defined(__STDC__) || defined(__cplusplus) || defined(_proto) || defined(c_plusplus)
#if defined(__cplusplus)
#define __MANGLE__	"C"
#else
#define __MANGLE__
#endif
#define __STDARG__
#define __PROTO__(x)	x
#define __OTORP__(x)
#define __PARAM__(n,o)	n
#if !defined(__STDC__) && !defined(__cplusplus)
#if !defined(c_plusplus)
#define const
#endif
#define signed
#define void		int
#define volatile
#define __V_		char
#else
#define __V_		void
#endif
#else
#define __PROTO__(x)	()
#define __OTORP__(x)	x
#define __PARAM__(n,o)	o
#define __MANGLE__
#define __V_		char
#define const
#define signed
#define void		int
#define volatile
#endif
#if defined(__cplusplus) || defined(c_plusplus)
#define __VARARG__	...
#else
#define __VARARG__
#endif
#if defined(__STDARG__)
#define __VA_START__(p,a)	va_start(p,a)
#else
#define __VA_START__(p,a)	va_start(p)
#endif
#endif
#include <ast.h>

#define	OFFSET		14
#define	ESCCODE		30

typedef struct
{
	Sfio_t*	fp;
	char*	end;
	int	count;
	int	peek;
	short	found;
	short	match;
	char	bigram1[(1<<(CHAR_BIT-1))];
	char	bigram2[(1<<(CHAR_BIT-1))];
	char	path[PATH_MAX];
	char	pattern[1];
} Find_t;

/*
 * return a fastfind stream handle for pattern
 */

__V_*
findopen __PARAM__((const char* pattern), (pattern)) __OTORP__(const char* pattern;){
	Find_t*	fp;
	char*		p;
	char*		s;
	char*		b;
	int		i;
	int			brace = 0;
	int			paren = 0;

	static char*		codes[] =
	{
		0,
		"/usr/local/share/lib/find/find.codes",
		"/usr/lib/find/find.codes"
	};

	if (!(fp = newof(0, Find_t, 1, 2 * (strlen(pattern) + 1)))) return(0);
	if (!codes[0]) codes[0] = getenv("FINDCODES");
	for (i = 0; i < elementsof(codes); i++)
		if (codes[i] && (fp->fp = sfopen(NiL, codes[i], "r")))
			break;
	if (!fp->fp)
	{
		free(fp);
		return(0);
	}
	for (i = 0; i < elementsof(fp->bigram1); i++) 
	{
		fp->bigram1[i] = sfgetc(fp->fp);
		fp->bigram2[i] = sfgetc(fp->fp);
	}

	/*
	 * extract last glob-free subpattern in name for fast pre-match
	 * prepend 0 for backwards match
	 */

	b = fp->pattern;
	p = s = (char*)pattern;
	for (;;)
	{
		switch (*b++ = *p++)
		{
		case 0:
			break;
		case '\\':
			s = p;
			if (!*p++) break;
			continue;
		case '[':
			if (!brace)
			{
				brace++;
				if (*p == ']') p++;
			}
			continue;
		case ']':
			if (brace)
			{
				brace--;
				s = p;
			}
			continue;
		case '(':
			if (!brace) paren++;
			continue;
		case ')':
			if (!brace && paren > 0 && !--paren) s = p;
			continue;
		case '|':
		case '&':
			if (!brace && !paren)
			{
				s = "";
				break;
			}
			continue;
		case '*':
		case '?':
			s = p;
			continue;
		default:
			continue;
		}
		break;
	}
	if (s != pattern) fp->match = 1;
	*b++ = 0;
	if (!*s) *b++ = '/';
	else while (i = *s++) *b++ = i;
	*b-- = 0;
	fp->end = b;
	fp->peek = sfgetc(fp->fp);
	return((__V_*)fp);
}

/*
 * close an open fastfind stream
 */

void
findclose __PARAM__((__V_* handle), (handle)) __OTORP__(__V_* handle;){
	Find_t*	fp = (Find_t*)handle;

	sfclose(fp->fp);
	free(fp);
}

/*
 * return the next fastfind path
 * 0 returned when list exhausted
 */

char*
findnext __PARAM__((__V_* handle), (handle)) __OTORP__(__V_* handle;){
	Find_t*	fp = (Find_t*)handle;
	char*		p;
	char*		q;
	char*		s;
	char*		b;
	int		c;
	int			n;

	for (c = fp->peek; c != EOF;)
	{
		if (c == ESCCODE)
		{
			if (sfread(fp->fp, (char*)&n, sizeof(n)) != sizeof(n)) break;
			c = n;
		}
		fp->count += c - OFFSET;
		for (p = fp->path + fp->count; (c = sfgetc(fp->fp)) > ESCCODE;)
			if (c & (1<<(CHAR_BIT-1)))
			{
				*p++ = fp->bigram1[c & ((1<<(CHAR_BIT-1))-1)];
				*p++ = fp->bigram2[c & ((1<<(CHAR_BIT-1))-1)];
			}
			else *p++ = c;
		*p-- = 0;
		b = fp->path;
		if (fp->found) fp->found = 0;
		else b += fp->count;
		for (s = p; s >= b; s--) 
			if (*s == *fp->end)
			{
				for (p = fp->end - 1, q = s - 1; *p; p--, q--)
					if (*q != *p)
						break;
				if (!*p)
				{
					fp->found = 1;
					if (!fp->match || strmatch(fp->path, fp->pattern))
					{
						fp->peek = c;
						return(fp->path);
					}
					break;
				}
			}
	}
	fp->peek = c;
	return(0);
}
